// Expression.cpp: implementation of the Expression class.
//
//////////////////////////////////////////////////////////////////////

#include "main.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Expression::Expression(String exp, BaseObject * obj, vector<String> &localNames)
{
	// Init all pointers to zero.
	
	cond = 0;
	expression = 0;
	increment = 0;
	init = 0;
	subExp = 0;
	elseExp = 0;

	// Decide what kind of expression this is.
	String word = exp.getWordAt(0);
	int pos = exp.getLastPos();

	if(exp.getNextChar(0) == '{')
	{
		type = EXP_BLOCK;
		String blockStr = exp.getCodeBlockAt(pos);
		StringList statements;

		blockStr.splitIgnoreAll(';', statements);
		for(int i = 0; i < statements.getSize(); i++)
		{
			String &str = statements[i];
			
			// We need to check if there is a block contained in this statement.
			// If so, the statement ends at a '}' intstead of a semicolon,
			// (unless it is an else statement), so we might need to parse
			// another expression following the '}' since it wouldn't be 
			// sperated by a semicolon.

			int count = str.countChar('}');
			if(count == 0)
			{
				if(str.getNextChar(0) != 0) // Check if this statement is empty
					block.push_back(new Expression(str, obj, localNames));
			} else {
				char * c = str;
				pos = strchr(c, '{')-c;
				
				String bl = str.getCodeBlockAt(pos);

				pos = bl.getLen() + pos + 2;
				

				String temp = str.getSubString(0, pos);
				if(temp.getNextChar(0) != 0)
					block.push_back(new Expression(temp, obj, localNames));
				
				temp = str.getSubString(pos+1, str.getLen());
				
				if(temp.getNextChar(0) != 0)
				{
					// Check if there are statements following the bracketed block

					if(temp.getWordAt(0) == "else")
					{
						// If there is an else at the begining of this, then this is 
						// probably a the else clause of an "if" expression. We put 
						// this in elseExp of the last expression
						block.back()->elseExp = new Expression(temp.getSubString(temp.getLastPos() + 4, temp.getLen()), obj, localNames);
					}
					else
					{
						// This is just a statement following the last.
						block.push_back(new Expression(temp, obj, localNames));
					}
				}
			}
		}

	} else if(word == "for")
	{
		type = EXP_FOR;
		String loop;
		loop = exp.getParenBlockAt(pos + 3);
		pos = exp.getLastPos() + loop.getLen() + 1;
		
		StringList loopItems;

		loop.split(';', loopItems);

		if(loopItems.getSize() != 3) throw LangError("Bad 'for' loop");

		// "For" statement has the structure:
		// for(init; cond; iter)

		init = new ExpressionNode(loopItems[0], obj, localNames);
		cond = new ExpressionNode(loopItems[1], obj, localNames);
		increment = new ExpressionNode(loopItems[2], obj, localNames);
		
		subExp = new Expression(exp.getSubString(pos, exp.getLen()), obj, localNames);

	} else if (word == "if") {

		type = EXP_IF;

		String condStr = exp.getParenBlockAt(pos + 2);

		cond = new ExpressionNode(condStr, obj, localNames);

		subExp = new Expression(exp.getSubString(condStr.getLen() + exp.getLastPos() + 1, exp.getLen()), obj, localNames);


	} else if (word == "while") {

		type = EXP_WHILE;

		String condStr = exp.getParenBlockAt(pos + 5);

		cond = new ExpressionNode(condStr, obj, localNames);

		subExp = new Expression(exp.getSubString(condStr.getLen() + exp.getLastPos() + 1, exp.getLen()), obj, localNames);

	} else if (word == "return") {

		type = EXP_RETURN;
		String temp = exp.getSubString(pos + 6, exp.getLen());
		expression = new ExpressionNode(temp, obj, localNames);

	} else if (word == "break") {
		
		type = EXP_BREAK;

		String endStr = exp.getSubString(pos + 5, exp.getLen());

		endStr.trimWhitespace();
		if(endStr != "")
			throw LangError("Expecting semicolon after 'break'");

	} else if (word == "continue") {
		
		type = EXP_CONTINUE;

		String endStr = exp.getSubString(pos + 8, exp.getLen());

		endStr.trimWhitespace();
		if(endStr != "")
			throw LangError("Expecting semicolon after 'continue'");

	} else if (word == "var") {

		// We need to check if this is an array declaration. If not,
		// treat it as a normal expression

		String name = exp.getWordAt(pos + 3);

		pos = exp.getLastPos() + name.getLen();

		if((pos < exp.getLen()) && (exp.getNextChar(pos) == '['))
		{
			type = EXP_ARRAY;
			// This is an array declaration
			for(int i=0; i<localNames.size(); i++)
			{
				if(localNames[i] == name)
					throw LangError("Variable already defined", __CURPOS__);
			}
			localNames.push_back(name);

			// The first element is the local variable index.
			dimensions.push_back(localNames.size()-1);

			while((pos < exp.getLen()) && (exp.getNextChar(pos) == '['))
			{
				String dim = exp.getBracketBlockAt(exp.getLastPos());
				dimensions.push_back(strtol(dim, 0, 10));
				pos = exp.getLastPos() + dim.getLen() + 1;
			}

		} else {
			// This is a regular declaration, evaluate normally
			type = EXP_NORMAL;
			expression = new ExpressionNode(exp, obj, localNames);
		}
	
	} else {
			type = EXP_NORMAL;
			expression = new ExpressionNode(exp, obj, localNames);		
	}
}

Expression::~Expression()
{
	if(expression) delete expression;
	if(cond) delete cond;
	if(increment) delete increment;
	if(init) delete init;
	if(subExp) delete subExp;
	if(elseExp) delete elseExp;
	for(int i=0; i<block.size(); i++)
	{
		delete block[i];
	}
}

Value Expression::evaluate(BaseObject *obj, vector<Value> &locals)
{
	switch(type)
	{
	case EXP_NORMAL:
		{
			return expression->evaluate(obj, locals);
			break;
		}
	case EXP_BLOCK:
		{
			for(int i=0; i<block.size(); i++)
			{
				Value ret = block[i]->evaluate(obj, locals);
				if(ret.flags & RETURN)
					return ret;
				if(ret.flags & BREAK)
					return ret;
				if(ret.flags & CONTINUE)
					return ret;
			}
			break;
		}
	case EXP_FOR:
		{
			
			for(init->evaluate(obj, locals); cond->evaluate(obj, locals).isTrue(); increment->evaluate(obj, locals))
			{
				Value ret = subExp->evaluate(obj, locals);
				if( (ret.flags & BREAK) || (ret.flags & RETURN) )
				{
					ret.flags &= ~BREAK;
					return ret; // If this is a break, this will be a NULL value
				}
				

				//c = cond->evaluate(obj, locals);
			}
			break;
			
		}
	case EXP_IF:
		{
			Value c = cond->evaluate(obj, locals);
			if(c.isTrue())
			{
				Value ret = subExp->evaluate(obj, locals);

				if((ret.flags & BREAK) || (ret.flags & RETURN))
				{
					return ret;
				}
				
			} else if (elseExp != 0)
			{
				Value ret = elseExp->evaluate(obj, locals);
				
				if((ret.flags & BREAK) || (ret.flags & RETURN))
				{
					return ret;
				}
			}
			break;
			
		}
	case EXP_WHILE:
		{
			Value c = cond->evaluate(obj, locals);
			while(c.isTrue())
			{
				Value ret = subExp->evaluate(obj, locals);
				if( (ret.flags & BREAK) || (ret.flags & RETURN) )
				{
					ret.flags &= ~BREAK;
					return ret; // If this is a break, this will be a NULL value
				}

				c = cond->evaluate(obj, locals);
			}
			break;
			
		}
	case EXP_RETURN:
		{
			Value ret = expression->evaluate(obj, locals);
			ret.flags |= RETURN;
			return ret;
		}
	case EXP_BREAK:
		{
			Value ret;
			ret.flags |= BREAK;
			return ret;
		}
	case EXP_CONTINUE:
		{
			Value ret;
			ret.flags |= CONTINUE;
			return ret;
		}
	case EXP_ARRAY:
		{
			// This is a local array creation

			// The first element of the dimensions
			// vector is the index of the variable
			Value &v = locals[dimensions[0]];

			// Recursively create array
			create(v, 1);
			
		}
	
	}
	return Value();
}

void Expression::create(Value &val, int dim)
{
	int size = dimensions[dim];
	val.createArray(size);
	dim ++;
	if(dim >= dimensions.size()) return;
	for(int i=0; i<size; i++)
	{
		create(*(val.getElement(i)), dim);
	}
}
