// ExpressionNode.cpp: implementation of the ExpressionNode class.
//
// 
//////////////////////////////////////////////////////////////////////

#include "main.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ExpressionNode::ExpressionNode(String &exp, BaseObject *obj, vector<String> &localNames)
{
	left = 0;
	right = 0;
	val = 0;
	create(exp, obj, localNames);
}

ExpressionNode::ExpressionNode(vector<ExpTerm> &list, BaseObject *obj, vector<String> &localNames)
{
	left = 0;
	right = 0;
	val = 0;
	create(list, obj, localNames);
}

void ExpressionNode::create(String &exp, BaseObject *obj, vector<String> &localNames)
{
	vector<ExpTerm> list;
	// First we need to split up the expression
	createTermList(exp, list, localNames);

	create(list, obj, localNames);

}

void ExpressionNode::create(vector<ExpTerm> &list, BaseObject *obj, vector<String> &localNames)
{
	int i; // Loop var

	// First we check if the list has only one element. 
	// If so, this node should be a leaf, unless the last 
	// term is a paranthesized term, or if the term contains
	// a unary oprerator
	if(list.size() == 1)
	{

		
		if(list[0].oper != 0)
		{

			// This is either a postfix increment or decrement

			// Postfix operators have to be dealt differently than prefix operators, which are dealt with below
			type = expOp;
			oper = list[0].oper;
			list[0].oper = 0;
			left = new ExpressionNode(list, obj, localNames);
		} else if(list[0].type == block)
		{
			// This is a paranethesized block
			create(list[0].str, obj, localNames);

		} else if (list[0].type == stringLiteral) 
		{
			val = new Value(list[0].str);
			type = expConst;
		} else if (list[0].type == numericLiteral) {

			int decimalCount = list[0].str.countChar('.');

			if(decimalCount == 0)
			{
				// This is an integer value
				val = new Value(strtol(list[0].str, 0, 10));
			} else if(decimalCount == 1) {
				val = new Value((float)strtod(list[0].str, 0));
			} else {
				throw LangError("Bad numeric constant", __FILE__, true);
			}

			type = expConst;

		} else if (list[0].type == symbol) {

			if(!list[0].noIndex)
			{

				// First we check if this is a local symbol
				for(i=0; i<localNames.size(); i++)
				{
					if(localNames[i] == list[0].str)
					{
						type = expLocal;
						localIndex = i;
						return;
					}
				}
						
				// Check if it is a member symbol
				
				for(i=0; i<obj->vars.nameList.size(); i++)
				{
					if(obj->vars.nameList[i] == list[0].str)
					{
						type = expMember;
						memberIndex = i;
						return;
					}
				}

			}
			// This is another type of symbol (either a function name or an object name)
			type = expSymbol;

			sym = new String(list[0].str);
			
		} else if(list[0].type  == paramList) {
			type = expParamList;
			pList = new vector<ExpressionNode *>;
						
			vector<String> strList;
			list[0].str.splitParamList(strList);
			list[0].str.trimWhitespace();
			if(list[0].str != "")
			{
				for(int i=0; i<strList.size(); i++)
				{
					pList->push_back(new ExpressionNode(strList[i], obj, localNames));
				}
			}
		} else if(list[0].type == arrayDimensions) {
			type = expParamList;
			pList = new vector<ExpressionNode *>;
						
			vector<String> strList;
			// The list cannot be empty, since an exception would have been thrown at String::getArrayDimensionBlockAt()
            int pos = 0;
			while(1)
			{
				if(pos >= list[0].str.getLen() || list[0].str.getNextChar(pos) != '[') break;

				strList.push_back(list[0].str.getBracketBlockAt(pos));
                pos = list[0].str.getLastPos() + strList.back().getLen() + 2;
			}

			//list[0].str.splitParamList(strList);
			//list[0].str.trimWhitespace();
			for(int i=0; i<strList.size(); i++)
			{
				pList->push_back(new ExpressionNode(strList[i], obj, localNames));
			}
			
		} else {
			throw LangError("Unexpected situation", __FILE__, true);
		}

		return;
	}

	// Since there is more than one element in the list, 
	// we know that this is going to be an internal node (aka operator)

	// We must search for the lowest precedence operator, since this will
	// be lower in the tree than any successive node. Also, we must reverse
	// associativity (left to right is parsed from right to left) so that
	// we are finding the lowest precedence operator.
	
	int curItem = 0;
	long num = (long)list.size();

	// Compound assignment operators are lowest 
	// precedence. right to left precedence
	

	for(i = 0; i < num; i++)
	{
		int op = list[i].oper;
		//   <<=, >>=, %=, *=, 
		//   /=, +=, -=, &=, 
		//   |=, ^=, =
		if(op == 'q' || op == 'Q' || op == 'U' || op == 't' ||
			op == 'd' || op == 'p' || op == 'm' || op == 'A' ||
			op == 'O' || op == 'X' || op == '=')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}

	// Logical OR "||", Left to right

	for(i = num - 1; i >= 0 ; i--)
	{
		int op = list[i].oper;

		if(op == 'o')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}

	// Logical AND "&&", Left to right

	for(i = num - 1; i >= 0 ; i--)
	{
		int op = list[i].oper;

		if(op == 'a')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}

	// Bitwise OR "|", Left to right

	for(i = num - 1; i >= 0 ; i--)
	{
		int op = list[i].oper;

		if(op == '|')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}

	// Bitwise exclusive OR "^", Left to right

	for(i = num - 1; i >= 0 ; i--)
	{
		int op = list[i].oper;

		if(op == '^')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}
	
	// Bitwise AND "&", Left to right

	for(i = num - 1; i >= 0 ; i--)
	{
		int op = list[i].oper;

		if(op == '&')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}

	// "==", "!=", Left to right

	for(i = num - 1; i >= 0 ; i--)
	{
		int op = list[i].oper;

		if(op == 'e' || op == 'n')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}

	// "<=", ">=", "<", ">", Left to right

	for(i = num - 1; i >= 0 ; i--)
	{
		int op = list[i].oper;

		if(op == 'L' || op == 'G' || op == '<' || op == '>')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}

	// ">>", "<<", Left to right

	for(i = num - 1; i >= 0 ; i--)
	{
		int op = list[i].oper;

		if(op == 'W' || op == 'w')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}

	// "+", "-", Left to right

	for(i = num - 1; i >= 0 ; i--)
	{
		int op = list[i].oper;

		if(op == '-' || op == '+')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}

	// "*", "/", "%", Left to right

	for(i = num - 1; i >= 0 ; i--)
	{
		int op = list[i].oper;

		if(op == '*' || op == '/' || op == '%')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}

	// Unary prefix operators, right to left precedence
	// unary negation '-', bitwise not '~', logical not '!',
	// pre-increment '++', pre-decrement '--', "new", "persistant", "delete", and dynamic array allocate "new var[]"
	

	for(i = 0; i < num; i++)
	{
		int op = list[i].oper;
		
		if(op == '@' || op == 'I' || op == 'D' || op == 'n' || op == '!' || op == '~' || op == 'c' || op == 'C' || op =='R')
		{
			makeOper(op, i, list, obj, localNames);
			return;
		}
	}

	// Special operators: '.' member access, '[]' indexing, '()' function call
	// and postfix increment/decrement

	for(i = num - 1; i >= 0; i--)
	{
		int op = list[i].oper;
		
		if(op == '.' || op == '[' || op == '(' || op == 'M' || op == 'P')
		{
			if(op == '.')
			{
				if((list.size()<=i) || (list[i+1].type != symbol))
				{
					throw LangError("Syntax Error", __CURPOS__);
				}

				list[i+1].noIndex = true; // We dont want the symbol being interpreted as a local or a member var
			}
			makeOper(op, i, list, obj, localNames);
			return;
		}
		
	}



	long s = (long)list.size();

	//char op = list[0].oper;

	throw LangError("Unsupported operator", __CURPOS__);



	
}

ExpressionNode::~ExpressionNode()
{
	if(type != expParamList)
	{
		if(left) delete left;
		if(right) delete right;
	} else {
		for(unsigned int i=0; i<pList->size(); i++)
		{
			delete (*pList)[i];
		}
		delete pList;
	}
	if(type == expConst) delete val;
	if(type == expSymbol) delete sym;
}

void ExpressionNode::createTermList(String &exp, vector<ExpTerm> &list, vector<String> &localNames)
{
	String word;
	char next;
	int pos=0;
	// The following bools are flags so that text following special operators can be handled differently on the next iteration
	bool wasIndex = false;
	bool wasFunction = false;
	bool wasUnary = false;
	bool wasDynArray = false;
	while(1)
	{
		if(pos >= exp.getLen()) return;

		next = exp.getNextChar(pos);
		pos = exp.getLastPos();

		if(next == 0) return;

		// Add an item to the back of the list
		list.push_back( ExpTerm() );
		ExpTerm &cur = list.back();


		// Check if the last operator was a unary operator
		if(wasUnary) goto getNextOperator;


		// First we have to determine the operand
		if(wasIndex) 
		{
			// If the last operator was an indexing operator,
			// we must be sure to put the entire block into the
			// operand. We then treat the contents as if it were
			// a normal parenthesized block.

			cur.str = exp.getBracketBlockAt(pos-1);
			pos += cur.str.getLen() + 1;
			cur.type = block;
			wasIndex = false;

/////// UNARY PREFIX OPERATORS /////////

			// For every unary prefix operator, the cur.str will be empty, since the operand
			// will be in the NEXT term
			//
			// [a | =] > [null | ++] > [i | null]

		} else if(wasFunction) {
			cur.type = paramList;
			cur.str = exp.getParenBlockAt(pos);
			pos += cur.str.getLen() + 2;
			wasFunction = false;
		} else if(wasDynArray) {
			cur.type = arrayDimensions;
			cur.str = exp.getArrayDimensionBlockAt(pos);
			pos += cur.str.getLen();
			wasDynArray = false;
		} else if(next == '!') {

			// This is a unary-NOT operator
			pos++;
			cur.type = prefixOp;
			cur.oper = '!';
			continue;

		} else if(next == '~') {

			// This is a bitwise complement operator
			pos++;
			cur.type = prefixOp;
			cur.oper = '~';
			continue;

		} else if(next == '+' && exp[pos+1] == '+') {

			// This is a prefix-increment operator
			pos+=2;
			cur.type = prefixOp;
			cur.oper = 'I';
			continue;

		} else if(next == '-' && exp[pos+1] == '-') {

			// This is a prefix-decrement operator
			pos+=2;
			cur.type = prefixOp;
			cur.oper = 'D';
			continue;



		} else if(next == '-' && !isdigit(exp[pos+1])) { // We know this isn't a decrement operator

			// This is a negation operator
			pos+=1;
			cur.type = prefixOp;
			cur.oper = 'n';
			continue;
////////// OPERANDS //////////

		} else if(next == '\"') {
			// This is a string literal
			int escCount=0;
			cur.str = exp.getLiteralAt(pos, escCount);
			pos += cur.str.getLen() + 2 + escCount;
			
			cur.type = stringLiteral;

		} else if ( isdigit(next) || next == '-' || next =='.') {
			// This is a numeric literal
			cur.str = exp.getNumericAt(pos);
			pos += cur.str.getLen();
			cur.type = numericLiteral;

		} else if( isalpha(next) ) {
			// This is a symbol.

			// First check if this is a keyword...

			// KEYWORD PROCESSING
			word = exp.getWordAt(pos);
			if( word == "var" )
			{
				pos = exp.getLastPos() + 3;
				cur.str = exp.getWordAt(pos);
				pos = exp.getLastPos() + cur.str.getLen();

				// Add the variable to the locals list.
				localNames.push_back(cur.str); 
				
				cur.type = symbol;
			} else if( word == "new" )
			{
				pos = exp.getLastPos() + 3; // Increment past 'new'
				if (exp.getWordAt(pos) == "var") // check if the next word is creating an array
				{
					pos = exp.getLastPos() + 3; // Increment past the 'var'

					// This is a dynamic array allocation
					cur.type = prefixOp;
					cur.oper = '@'; 

					wasDynArray = true;

					// An example of an ExpTerm list with a dyn array alloc
					//
					// 
					// [a | =] > [null | @] > ["[432][3]" | null]
					//

				} else {
					cur.type = prefixOp;
					cur.oper = 'c';
				}
				continue;
			} else if( word == "persistant" )
			{
				pos = exp.getLastPos() + 10;

				cur.type = prefixOp;
				cur.oper = 'C';
				continue;
			} else if( word == "delete" )
			{
				pos = exp.getLastPos() + 6;

				cur.type = prefixOp;
				cur.oper = 'R';
				continue;
			} else {

				// ... otherwise it is a regular symbol.
				cur.str = exp.getWordAt(pos);
				pos += cur.str.getLen();
				cur.type = symbol;
			}

		} else if( next == '(' ) {
			// This is a paranethisized block
			cur.str = exp.getParenBlockAt(pos);
			pos += cur.str.getLen() + 2;
			cur.type = block;			
		}

////////// OPERATORS /////////////

getNextOperator:  // Used for unary operators

		if(pos >= exp.getLen() ) return;
		next = exp.getNextChar(pos);
		pos = exp.getLastPos() + 1;
		
		if(next == 0) return;

		if(next == '.')
		{
			cur.oper = '.'; // "."
		} else if (next == '<')
		{
			next = exp[pos];
			if(next == '=')  
			{
				pos ++;
				cur.oper = 'L';  // "<="
			} else if (next == '<') {
				pos ++;
				next = exp[pos];
				if (next == '=')
				{
					pos ++;
					cur.oper = 'q'; // "<<="
				} else {
					cur.oper = 'w'; // "<<"
				}
			} else {
				cur.oper = '<'; // "<"
			}
		} else if (next == '>')
		{
			next = exp[pos];
			if(next == '=')  
			{
				pos ++;
				cur.oper = 'G';  // ">="
			} else if (next == '>') {
				pos ++;
				next = exp[pos];
				if (next == '=')
				{
					pos ++;
					cur.oper = 'Q'; // ">>="
				} else {
					cur.oper = 'W'; // ">>"
				}
			} else {
				cur.oper = '>'; // ">"
			}
		} else if (next == '=') {
			next = exp[pos];
			if(next == '=')
			{
				pos ++;
				cur.oper = 'e'; // "=="
			} else {
				cur.oper = '='; // "="
			}
		} else if (next == '%') {
			next = exp[pos];
			if(next == '=')
			{
				pos ++;
				cur.oper = 'U'; // "%="
			} else {
				cur.oper = '%'; // "%"
			}
		} else if (next == '*') {
			next = exp[pos];
			if(next == '=')
			{
				pos ++;
				cur.oper = 't'; // "*="
			} else {
				cur.oper = '*'; // "*"
			}
		} else if (next == '/') {
			next = exp[pos];
			if(next == '=')
			{
				pos ++;
				cur.oper = 'd'; // "/="
			} else {
				cur.oper = '/'; // "/"
			}
		} else if (next == '+') {
			next = exp[pos];
			if(next == '=')
			{
				pos ++;
				cur.oper = 'p'; // "+="
			} else if (next == '+')	{
				pos ++;
				cur.oper = 'P'; // "++"   Unary postfix operator
				wasUnary = true;
			} else {
				cur.oper = '+'; // "+"
			}
		} else if (next == '-') {
			next = exp[pos];
			if(next == '=')
			{
				pos ++;
				cur.oper = 'm'; // "-="
			} else if (next == '-')	{
				pos ++;
				cur.oper = 'M'; // "--"    Unary postfix operator
				wasUnary = true;
			} else {
				cur.oper = '-'; // "-"
			}
		} else if (next == '&')	{
			next = exp[pos];
			if(next == '=')  
			{
				pos ++;
				cur.oper = 'A';  // "&="
			} else if (next == '&') {
				pos ++;
				next = exp[pos];	
				if (next == '=')
				{
					pos ++;
					cur.oper = 'y'; // "&&="
				} else {
					cur.oper = 'a'; // "&&"
				}
			} else {
				cur.oper = '&'; // "&"
			}
		} else if (next == '|')	{
			next = exp[pos];
			if(next == '=')  
			{
				pos ++;
				cur.oper = 'O';  // "|="
			} else if (next == '|') {
				pos ++;
				next = exp[pos];
				if (next == '=')
				{
					pos ++;
					cur.oper = 'Y'; // "||="
				} else {
					cur.oper = 'o'; // "||"
				}
			} else {
				cur.oper = '|'; // "|"
			}
		} else if (next == '^') {
			next = exp[pos];
			if(next == '=')
			{
				pos ++;
				cur.oper = 'X'; // "^="
			} else {
				cur.oper = '^'; // "^"
			}
		} else if (next == '[') {
			// This one is a bit odd since as an operator, the square
			// brackets come before and after the second operand. We
			// make this a special case and set a flag.

			cur.oper = '['; // "[]", index into array
			wasIndex = true;
		} else if (next == '!' && exp[pos] == '=') {
			cur.oper = 'N'; // "!="
			
		} else if (next == '(') {
			cur.oper = '('; // Function call
			pos --;
			wasFunction = true;

		

		} else {
			throw LangError("Syntax error", __CURPOS__);
		}


		
	}

}


// Function makeOper()

// This takes a list and splits it into two parts, in the process making this expression node an
// internal node. This means that it is an operator

// Binary operator, expression 7 - 5 + 6 * 9
// Lowest precedence operator is +, thus split = 2
//
//    1         2         3         4
// [7 | -] > [5 | +] > [6 | *] > [9 | null]
//                   ^
//                   |
//                This is where the split occurs
//
//
// Expression node now is a '+' operator.
//
// [7 | -] > [5 | null]     [6 | *] > [9 | null]
//                    \     /
//                      '+'


// Unary operator
// Split = 1
//
//        1                 2
//  [null | new] > [myClass | null]
//               ^
//               |
//           Split is here
//
// 
// Expression is now a new operator
//
//
//           null     [myClass | null]
//              \     /
//                new

// Not counting special operators, unary operators have the highest precidence, and are always
// evaluated left to right, thus we can assume that if this is a unary operator, there will be
// no elements in the left list. You can see this in this example:
//
// [null | ++] > [null | --] > [a | null]
//
// Lowest precedence is the  ++, thus this becomes:
//
//            null  [null | --] > [a | null] 
//               \  /
//                ++

// This function makes this into an internal operator node
void ExpressionNode::makeOper(char op, int split, vector<ExpTerm> &list, BaseObject *obj, vector<String> &localNames)
{
	type = expOp;
	oper = op;
	vector<ExpTerm> rightList; // Right side of the split

	for(unsigned int i = split + 1; i < list.size(); i++)
	{
		rightList.push_back(list[i]);
	}

	list.resize(split + 1);
	list.back().oper = 0;
	if((list.back().str == "") && (list.back().type != paramList))
	{
		// This will occur for an operator following a unary operator
		list.resize(split);
	}
	int l = (int)list.size();
	int r = (int)rightList.size();

	if(list.size()>0) left = new ExpressionNode(list, obj, localNames);
	if(rightList.size()>0) right = new ExpressionNode(rightList, obj, localNames);
}

BaseObject *ExpressionNode::getObject(BaseObject *obj, vector<Value> &locals)
{
	switch (type)
	{
	case expConst:
		{
			throw LangError("Syntax error", __CURPOS__);
		}
	case expLocal:
		{
			Value &v = locals[localIndex];
			if(v.type != OBJECT)
				throw LangError("Non-class member access", __CURPOS__);
			return v.obj;
		}
	case expMember:
		{
			Value &v = obj->vars.valList[memberIndex];
			if(v.type != OBJECT)
				throw LangError("Non-class member access", __CURPOS__);
			return v.obj;
		}
	case expSymbol:
		{
			return obj->script->getObject(*sym);
		}
	}

	Value v = evaluate(obj, locals);
	if(v.type != OBJECT)
		throw LangError("Non-class member access", __CURPOS__);
	return v.obj;
}

Value ExpressionNode::evaluate(BaseObject *obj, vector<Value> &locals)
{
	switch (type)
	{
	case expConst:
		return *val;

	case expLocal:
		return locals[localIndex];

	case expMember:
		return obj->vars.valList[memberIndex];

	case expSymbol:
		// If we are being asked to evaluate a symbol, then it must be
		// an object

		return Value(obj->script->getObject(*sym));
	}


	// We know that this is an operator now.

	switch (oper)
	{
	case '=': // Assignment
		{
			Value &lvalue = left->getLValue(obj, locals);
			lvalue.assign(right->evaluate(obj,locals));
			return lvalue;

		}
////// Arithmetic //////
	case '+':
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opAdd(&rvalue);
			return lvalue;
		}
	case '-':
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opSub(&rvalue);
			return lvalue;
		}
	case '*':
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opMult(&rvalue);
			return lvalue;
		}
	case '/':
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opDiv(&rvalue);
			return lvalue;
		}
	case '%':
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opModulo(&rvalue);
			return lvalue;
		}
////// Comparison operators //////
	case '<':
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opLess(&rvalue);
			return lvalue;
		}
	case '>':
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opGreater(&rvalue);
			return lvalue;
		}
	case 'L': // <=
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opLessOrEqual(&rvalue);
			return lvalue;
		}
	case 'G': // >=
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opGreaterOrEqual(&rvalue);
			return lvalue;
		}
	case 'e': // ==
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opEquality(&rvalue);
			return lvalue;
		}
////// Increment/Decrement //////
	case 'P': //"++"   Unary postfix operator
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value ret(lvalue);
			lvalue.inc();
			return ret; // Return the pre-incremented value
		}
	case 'M': //"--"   Unary postfix operator
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value ret(lvalue);
			lvalue.dec();
			return ret; // Return the pre-decremented value
		}
	case 'I': //"++"   Unary prefix operator
		{
			Value &lvalue = right->getLValue(obj, locals);
			lvalue.inc();
			return lvalue; // Return the post-incremented value
		}
	case 'D': //"--"   Unary prefix operator
		{
			Value &lvalue = right->getLValue(obj, locals);
			lvalue.dec();
			return lvalue; // Return the post-decremented value
		}
	case 'n': // "-" Unary negation operator
		{
			Value lvalue = right->evaluate(obj, locals);
			lvalue.negate();
			return lvalue;
		}
	case '!':
		{
			Value lvalue = right->evaluate(obj, locals);
			if(lvalue.isTrue()) 
			{
				lvalue.clear();
				lvalue.i = 0;
			} else {
				lvalue.clear();
				lvalue.i = 1;
			}
			lvalue.type = INTEGER;
			return lvalue;
		}
////// Creation /////
	case 'c': // "new"
		{
			BaseObject *base = right->getObject(obj, locals);
			if(!base)
			{

			} else {
				while((base->name=="") && (base->base))
				{
					base = base->base;
				}
				Object *n = new Object(obj->script);
				n->create(base);
				obj->script->addObject(n);

				return Value(n);
			}

		}
	case 'C': // "persistant"
		{
			BaseObject *base = right->getObject(obj, locals);
			while((base->name=="") && (base->base))
			{
				base = base->base;
			}
			Object *n = new Object(obj->script);
			n->create(base);
			n->isPersistant = true;
			obj->script->addObject(n);

			return Value(n);

		}
	case 'R': // "delete"
		{
			BaseObject *delObj = right->getObject(obj, locals);
			obj->script->deleteObject(delObj);
			return Value();
		}
	case '@': // "new var[]", dynamic array creation
		{
			if(right->type != expParamList) throw LangError("Unexpected situation, looking for parameter list");
			
			vector<ExpressionNode *> &e = *(right->pList);
			vector<int> sizes;
            for(int i = 0; i < e.size(); i++)
			{
				Value v = e[i]->evaluate(obj, locals);
				v.convInt();
				sizes.push_back(v.i);
			}
			Value ret;
			ret.createArray(sizes);
			return ret;
			
		}
////// Shift operators //////
	case 'w': // "<<"
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opShiftLeft(&rvalue);
			return lvalue;
		}
	case 'W': // ">>"
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opShiftRight(&rvalue);
			return lvalue;
		}
////// Compound assignment operators //////
	case 'q': // "<<="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opShiftLeft(&rvalue);
			return lvalue;
		}
	case 'Q': // ">>="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opShiftRight(&rvalue);
			return lvalue;
		}
	case 'U': // "%="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opModulo(&rvalue);
			return lvalue;
		}
	case 'p': // "+="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opAdd(&rvalue);
			
			return lvalue;
		}
	case 'm': // "-="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opSub(&rvalue);
			return lvalue;
		}
	case 'd': // "/="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opDiv(&rvalue);
			return lvalue;
		}
	case 't': // "*="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opMult(&rvalue);
			return lvalue;
		}
	case 'A': // "&="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opBitwiseAnd(&rvalue);
			return lvalue;
		}
	case 'O': // "|="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opBitwiseOr(&rvalue);
			return lvalue;
		}
	case 'y': // "||="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opLogicalOr(&rvalue);
			return lvalue;
		}
	case 'Y': // "&&="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opLogicalAnd(&rvalue);
			return lvalue;
		}
	case 'X': // "^="
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opBitwiseXor(&rvalue);
			return lvalue;
		}
	case 'a': // "&&"
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opLogicalAnd(&rvalue);
			return lvalue;
		}
	case 'o': // "||"
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opLogicalOr(&rvalue);
			return lvalue;
		}
	case '&': // "&"
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opBitwiseAnd(&rvalue);
			return lvalue;
		}
	case '|': // "|"
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opBitwiseOr(&rvalue);
			return lvalue;
		}
	case '^': // "^"
		{
			Value lvalue = left->evaluate(obj, locals);
			Value rvalue = right->evaluate(obj, locals);
			lvalue.opBitwiseXor(&rvalue);
			return lvalue;
		}
	case '.':
		{
			// The left side must evaluate to an object.

			// Since a value is being requested, the right side must
			// be referring to the name of a member variable in the left side

			BaseObject *o = left->getObject(obj, locals);
			if(right->type != expSymbol)
				throw LangError("Expecting member function or variable", __CURPOS__);
			return o->vars[*(right->sym)];
		}
	case '(':
		{
			// Left must resolve to a function, right must resolve to a parameter list
			
			//Function *func;
			BaseObject *context;
			vector<Value> parm;

			if(right->type != expParamList) throw LangError("Unexpected situation, looking for parameter list");
			
			vector<ExpressionNode *> &e = *(right->pList);
			int size = (int)e.size();
			parm.resize(size);
			for(int i=0; i<size; i++)
			{
				parm[i] = e[i]->evaluate(obj, locals);
			}

			if(left->type == expOp)
			{
				// This should be a dot operator, accessing a member function
				// of another class. This changes the context of the function
				// call
				
				if(left->oper != '.') throw LangError("Bad function call", __CURPOS__);
				if(left->right->type != expSymbol) throw LangError("Bad function call", __CURPOS__);

				context = left->left->getObject(obj, locals);
				if(!context) throw LangError("Object not found", __CURPOS__);
				return context->callFunction(*(left->right->sym), parm);

			} else if(left->type == expSymbol) {
				return obj->callFunction(*(left->sym), parm);
			} else {
				throw LangError("Syntax error", __CURPOS__);
			}
		}
	case '[':
		{
			Value &lvalue = left->getLValue(obj, locals);
			Value index = right->evaluate(obj, locals);

			if(index.type != INTEGER)
				throw LangError("Non-integer array index", __CURPOS__);

			return *(lvalue.getElement(index.i));

		}
	default:
		{
			throw LangError("Unexpected situation (unrecognized operator)", __CURPOS__);
		}

	}
	return Value();
}

Value& ExpressionNode::getLValue(BaseObject *obj, vector<Value> &locals)
{
	switch (type)
	{
	case expLocal:
		return locals[localIndex];
	case expMember:
		return obj->vars.valList[memberIndex];
	case expOp:
		{
			// The only operators that return an lvalue are the dot operator and index operator
			if(oper=='.')
			{
				BaseObject *o = left->getObject(obj, locals);
				if(right->type != expSymbol)
					throw LangError("Expecting member function or variable", __CURPOS__);
				return o->vars[*(right->sym)];

			} else if(oper=='[')
			{
				Value &lvalue = left->getLValue(obj, locals);
				Value index = right->evaluate(obj, locals);

				if(index.type != INTEGER)
					throw LangError("Non-integer array index", __CURPOS__);

				return *(lvalue.getElement(index.i));

			} else {
				throw LangError("Expecting LValue", __CURPOS__);
			}
		}
	default:
		throw LangError ("Unexpected Situation", __CURPOS__);
	}
}