// Script.cpp: implementation of the Script class.
//
//////////////////////////////////////////////////////////////////////

#include "main.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Script::Script(const char *filename) //: objList(MAXOBJ, 0)
{ 
	sys=0;
	isLoaded=false;
	loadScript(filename);
}

Script::Script() //: objList(MAXOBJ, 0)
{
	isLoaded=false;
	sys=0;
}

Script::~Script()
{
	for(unsigned int i=0; i<objList.size(); i++)
	{
		if(objList[i]) delete objList[i];
	}
	
}

//////////////////////////////////////////////////////////////////////
// Member functions
//////////////////////////////////////////////////////////////////////

// This is the main function used to load scripts. At some point, it should allow
// more than one script to be loaded.
void Script::loadScript(const char * filename)
{

	
	String text;
	
	// Load the file and do a little text processing
	text=preParse(filename).implode("\n");
	removeTabs(text);

	// Create the objects
	parse(text);

	for(long i = (long)objList.size() - 1; i>=0; i--)
	{
		if(objList[i]) objList[i]->construct(objList[i]);
	}

	isLoaded = true;	
}

// preParse does three things, it loads the file, removes C-style comments, 
// and processes #include's
StringList Script::preParse(const char *filename)
{
	long i;
	long lines;
	String text;
	StringList list;

	text.loadFile(filename);

	// First, split the file into lines
	text.split('\n', list);

	lines = list.getSize();

	for(i=0; i<list.getSize(); i++)
	{
		// str becomes an alias for the current string (list[i])
		String &str=list[i];
		int len=str.getLen();
		char *strp=str;
		bool isLit=false;
		for(int j=0; j<len; j++)
		{
			// We need to keep track of literals. isLit is true if the current
			// position is inside a literal string.
			if(strp[j]=='\"')
			{
			if(!isLit) 
				isLit=true;
			else if(j!=0 && strp[j-1]!='\\') 
				isLit=false;
			}
			
			// If we come to two '/' then we know the rest of the line is a comment
			if(!isLit && strp[j]=='/' && strp[j+1]=='/')
			{
				// We just make it whitespace, and leave it up to the trimWhitespace()
				// function to take out the blank space
				for(int k=j; k<len; k++)
					strp[k]=' ';
				break;
			}
		}

		// The removed comments leave large runs of spaces, which needs to be gotten rid of
		str.trimWhitespace();

		// TODO: Check if the whitespace is stripped before this point. If it isnt,
		// then it would be required that #includes start at the beginning of the line
		if(list[i][0]=='#') 
		{
			String word;
			word=str.getWordAt(1);
			// We could add more preParse directives here, such as #define
			if(word == "include")
			{
				int j;
				list.insertList(preParse(str.getLiteralAt(8, j)),i+1);
				list.deleteString(i);
			}

		}
	}

	// This is of course, very inefficient... not too important tho
	return list;
}

// This replaces all tabs in a string with spaces. This might be better suited as
// a String member function.
void Script::removeTabs(String &str)
{
	int pos;
	bool isLit=false;
	for(pos=0; pos<str.getLen(); pos++)
	{
		if(str[pos]=='\"')
		{
			if(!isLit) 
				isLit=true;
			else if(pos!=0 && str[pos-1]!='\\') 
				isLit=false;
		}
		if(!isLit && str[pos]=='\t') str[pos]=' ';
	}
}

// Script::parse() is one of the most important functions in this class. It must
// parse the object definitions from the files. It doesn't actually have to do
// anythine with the bodies of the objects, it supplies this information to the
// Objects, which handle the further creation.
//
// The function is divided into two parts.
//
// Part 1: Lists are created holding the bodies, base, and name
//
// Part 2: Create the objects based on the lists created in part 1
void Script::parse(String &str)
{
	

	long pos=0;
	String word;
	
	// These lists contain information about the objects, filled in by part 1 of
	// this function, then used by part 2 to create the objects.

	vector<String> baseName;
	vector<String> name;
	vector<String> body;
	vector<int> lineNum;

	bool isFinal = false;
	bool isBase = false;

	//int curObj = 0; // Current object index

////////////// PART 1 /////////////////
	
	while(pos < str.getLen())
	{
		
		word = str.getWordAt(pos);
		pos = str.getLastPos();
		// No more words are left, break out
		if(word == "") break;

		if(word == "final") 
		{
			isFinal = true;
			pos+=5; // Advance position past "final"
		} else if(word == "base") 
		{
			isBase = true;
			pos+=4; // Advance position past "base"
		} else if(word == "class")
		{
			Object *obj = new Object(this);
			obj->script = this; // Set script pointer
			pos+=5; // advance the position past "class"

			// Get the objects name
			word = str.getWordAt(pos);
			pos=str.getLastPos() + word.getLen();

			// Check to see if the name is a reserved word, or has already been registered
			if(int err = checkName(word))
			{
				if(err == 1)
				{
					int line = str.getLine(pos);
					throw LangError(String("Attempted use of reserved word as class name in line ") + String::toString(line) + ", use of '" + word + "'");
				} else {
					int line = str.getLine(pos);
					throw LangError(String("Attempt to redefine class") + word +  " in line " + String::toString(line));
				}
			}

			obj->name=word;
			name.push_back(word);
			// Get the next character to check for inheritance
			char next=str.getNextChar(pos);
			pos=str.getLastPos();

			if(next==':')
			{
				long curObj = (long)baseName.size();
				baseName.push_back( str.getWordAt(pos + 1) );
				
				pos=str.getLastPos() + baseName.back().getLen();

				// Set up the next character for the body parse
				next=str.getNextChar(pos);
				pos=str.getLastPos();
			} else {
				baseName.push_back( "" );
			}
			
			// The next character SHOULD be an '{'
			if(next='{') {
				// Check to make sure the code block is okay
				try 
				{
					word = str.getCodeBlockAt(pos);
				} catch (LangError e)
				{
					e.errText = "Reached end of file looking for closing brace '}'";
					throw e;
				}
				//obj->create(word);
				body.push_back( word );
				lineNum.push_back( str.getLine( str.getLastPos() ) );
				obj->isBase = isBase;
				obj->isFinal = isFinal;
				isFinal = false;
				isBase = false;
				pos = str.getLastPos() + word.getLen() + 1;
			} else {
				delete obj;
				int line = str.getLine(pos);
				String errMsg = String("Syntax error in line ") + String::toString(line) + " while looking for '{' or ':', found '" + String::toString(next) + "'";

				throw LangError(errMsg, "Script::parse(String &str)", true);
			}
			
			// Add this object
			objList.push_back( obj );
			
		} else {
			int line = str.getLine(pos);
			String errMsg = String("Syntax error in line ") + String::toString(line) + " while looking for 'base' or 'class', found '" + word + "'";
			throw LangError(errMsg, "Script::parse(String &str)", true);
		}
	}

////////////// PART 2 /////////////////

	unsigned int i;
	// Iterate list and assign base pointers
	for(i = 0; i < name.size(); i++)
	{
		getObject(name[i])->base = getObject(baseName[i]);
	}
	

	// This must be seperate from the above loop so that all objects are resolved
	// before the first object is created
	bool allCreated = false;
	while(!allCreated)
	{
		allCreated = true;
		for(i = 0; i < name.size(); i++)
		{
			Object * obj = (Object *)getObject(name[i]);
			if(!(obj->isCreated))
			{
				if(!(obj->base) || obj->base->isCreated)
				{
					obj->create(body[i], lineNum[i]);
				} else {
					allCreated = false;
				}
			}
		}
	}
}

BaseObject * Script::getObject(const String &name)
{
	for(unsigned int i = 0; i < objList.size(); i++)
	{
		if(objList[i] && objList[i]->name == name) return objList[i];
	}
	return 0;
}


int Script::checkName(String &name)
{
	
	// Check if this is a reserved word
	if( name == "if" || name == "for" || name == "return" ||
		name == "class" || name == "do" || name == "while" || 
		name == "var" || name == "else" || name == "function" ||
		name == "final" || name == "base" || name == "new" || 
		name == "delete")

		return 1; // Reserved word used
	
	if(getObject(name)!=0) 
		return 2; // Name already defined
	

	return 0;
	
}


void Script::addObject(BaseObject *object)
{
	object->script=this;
	//for(int i = 0; i < objList.getSize(); i++)
	for(unsigned int i = 0; i < objList.size(); i++)
	{
		if(!objList[i])
		{
			objList[i] = object;
			return;
		}
	}
	
	objList.push_back(object);

	
	//throw LangError("No object space left!", __FILE__);
}

void Script::registerDefSystem()
{
	addObject(new DefaultSystem());
}


/*
void Script::getAllObjects(List<BaseObject *> &list)
{
	// REIMPLEMENT

	
	List<BaseObject *> temp(MAXOBJ, 0);
	int j=0;
	int i;
	for(i=0; i<MAXOBJ; i++)
	{
		if(objList[i] && !(objList[i]->isBase))
		{
			temp[j]=objList[i];
			j++;
		}
	} 
	list.resize(j);
	for(i = 0; i < j; i++)
	{
		list[i]=temp[i];
	}
	
	
}
*/

/*
void Script::getAllSubclass(List<BaseObject *> &list, BaseObject *base)
{
	
	List<BaseObject *> temp(MAXOBJ, 0);
	int j=0;
	int i;
	for(i=0; i<MAXOBJ; i++)
	{
		if(objList[i] && !(objList[i]->isBase) && objList[i]->isSubClass(base))
		{
			temp[j]=objList[i];
			j++;
		}
	} 
	list.resize(j);
	for(i=0; i<j; i++)
	{
		list[i]=temp[i];
	}
	
}
*/

BaseObject * Script::getSystem()
{
	if(!sys) sys=getObject("system");
		return sys;
	return 0;
}

void Script::flush()
{
//	env.flush();
}

void Script::garbageCollect()
{
	
	for(unsigned int i=0; i<objList.size(); i++)
	{
		// Only nameless classes are affected by garbage collection
		if(objList[i] && objList[i]->isPersistant == false && objList[i]->name == "")
		{
			if(objList[i]->refList.isEmpty())
			{
				delete objList[i];
				objList[i] = 0;
			}
		}
	}

}

void Script::deleteObject(BaseObject *obj)
{
	for(unsigned int i=0; i<objList.size(); i++)
	{
		if(objList[i] == obj)
		{
			delete obj;
			objList[i] = 0;
		}
	}
}
